'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var inside = require('is-path-inside');
var _ = require('lodash');
var path = require('upath');
var fs = require('fs');
var staticCache = require('koa-static-cache');
var decorator = require('@midwayjs/decorator');
var superjson = require('superjson');
var consola = require('consola');
var asynchronousLocalStorage = require('asynchronous-local-storage');
var tslib = require('tslib');
var parseArgs = require('fn-args');
var esModuleLexer = require('es-module-lexer');
var chalk = require('chalk');
var createJITI = require('jiti');
var pkgDir = require('pkg-dir');
var util = require('util');
var globby = require('globby');
var path$1 = require('path');
var core = require('@midwayjs/core');
require('axios');
var request_dist_midwayjsHooksCoreRequest = require('../request/dist/midwayjs-hooks-core-request.cjs.prod.js');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var inside__default = /*#__PURE__*/_interopDefault(inside);
var ___default = /*#__PURE__*/_interopDefault(_);
var path__default = /*#__PURE__*/_interopDefault(path);
var staticCache__default = /*#__PURE__*/_interopDefault(staticCache);
var superjson__default = /*#__PURE__*/_interopDefault(superjson);
var consola__default = /*#__PURE__*/_interopDefault(consola);
var parseArgs__default = /*#__PURE__*/_interopDefault(parseArgs);
var chalk__default = /*#__PURE__*/_interopDefault(chalk);
var createJITI__default = /*#__PURE__*/_interopDefault(createJITI);
var path__default$1 = /*#__PURE__*/_interopDefault(path$1);

class FileRouter {
  constructor(options) {
    this.root = void 0;
    this.projectConfig = void 0;
    this.useSourceFile = void 0;
    const {
      root,
      projectConfig,
      useSourceFile
    } = options;
    this.root = root;
    this.projectConfig = projectConfig;
    this.useSourceFile = useSourceFile;
  }

  get source() {
    if (this.useSourceFile) {
      return path.join(this.root, this.projectConfig.source);
    }

    return path.join(this.root, this.projectConfig.build.outDir);
  } // src/apis/lambda


  getApiDirectory(baseDir) {
    return path.join(this.source, baseDir);
  }

  getRoute(file) {
    const {
      routes
    } = this.projectConfig;
    return routes.find(route => {
      const dir = this.getApiDirectory(route.baseDir);
      return this.inside(file, dir);
    });
  }

  getGatewayByRoute(route) {
    const gateway = this.projectConfig.gateway.find(adapter => adapter.is(route));

    if (!gateway) {
      throw new Error(`Can't find the correct gateway adapter, please check if midway.config.ts is correct`);
    }

    return gateway;
  }

  isApiFile(file) {
    if (file.endsWith('.test.ts') || file.endsWith('.test.js')) {
      return false;
    }

    if (path.extname(file) !== '.ts' && path.extname(file) !== '.js') {
      return false;
    }

    const route = this.getRoute(file);
    return !!route;
  }

  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = path.relative(this.source, file); // src/apis/lambda/index.ts -> apis-lambda-index

    const id = _.kebabCase(path.removeExt(relativePath, path.extname(relativePath)));
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('');
    return name.toLowerCase();
  }

  inside(child, parent) {
    return inside__default['default'](path.toUnix(child), path.toUnix(parent));
  }

}

/**
 * @private
 * private api, may change without notice.
 * Use asynchronous-local-storage due to serverless environment does not support node.js 12.17.0
 */
const als = {
  getStore(key) {
    return asynchronousLocalStorage.als.get(key);
  },

  run(ctx, callback) {
    return new Promise(resolve => {
      asynchronousLocalStorage.als.runWith(async () => {
        resolve(await callback());
      }, ctx);
    });
  }

};
function useContext() {
  return als.getStore('ctx');
}
function createFunctionContainer(options) {
  const {
    fn,
    functionId,
    parseArgs,
    handlerDecorators = [],
    classDecorators = [],
    isHTTP
  } = options;
  let FunctionContainer = class FunctionContainer {
    constructor() {
      this.ctx = void 0;
    }

    async handler(...handlerArgs) {
      const args = parseArgs(this.ctx, ...handlerArgs);
      return isHTTP ? await fn(...args) : await als.run({
        ctx: this.ctx
      }, async () => await fn(...args));
    }

  };

  tslib.__decorate([decorator.Inject()], FunctionContainer.prototype, 'ctx', void 0);

  tslib.__decorate([...handlerDecorators], FunctionContainer.prototype, 'handler', null);

  FunctionContainer = tslib.__decorate([decorator.Provide(functionId), ...classDecorators], FunctionContainer);
  return FunctionContainer;
}

function isDevelopment() {
  if (process.env.MIDWAY_TS_MODE === 'true' ||
  /* istanbul ignore next */
          "production" === 'test' ||
  /* istanbul ignore next */
          "production" === 'development') {
    return true;
  }
  /* istanbul ignore next */


  return false;
}
function useHooksMiddleware(fn) {
  return (...args) => {
    /**
     * Hooks middleware
     * const middleware = (next) => { const ctx = useContext() }
     */
    if (parseArgs__default['default'](fn).length === 1) {
      const next = _.last(args);
      return fn(next);
    }

    return fn(...args);
  };
}
function formatCode(code) {
  try {
    const prettier = require('prettier');

    return prettier.format(code, {
      semi: true,
      singleQuote: true,
      parser: 'babel'
    });
  } catch {
    return code;
  }
}

async function createHTTPApiClient(file, code, router) {
  await esModuleLexer.init;
  const [, exports] = esModuleLexer.parse(code);

  if (exports.length === 0) {
    return null;
  }

  const baseUrl = router.getBaseUrl(file);
  const {
    superjson,
    request: {
      client
    }
  } = router.projectConfig;
  const funcs = exports.map(name => {
    const isExportDefault = name === 'default';
    return {
      url: getUrl(baseUrl, name),
      meta: {
        superjson
      },
      functionId: router.getFunctionId(file, name, isExportDefault),
      isExportDefault: isExportDefault,
      name: isExportDefault ? '$default' : name
    };
  });
  return formatCode(buildClient$1(client, funcs));
}

function buildClient$1(client, funcs) {
  const buildFunction = func => {
    return `
      export ${func.isExportDefault ? 'default' : ''} function ${func.name} (...args) {
        return request({
          url: '${func.url}',
          method: args.length === 0 ? 'GET' : 'POST',
          data: { args },
          meta: ${JSON.stringify(func.meta, null, 2)}
        })
      }
    `;
  };

  const functions = funcs.map(buildFunction).join('\n');
  return `
    import { request } from '${client}';
    ${functions}
  `;
}

function getUrl(baseUrl, name) {
  if (name === 'default') {
    return baseUrl;
  }

  if (baseUrl.endsWith('/')) {
    return `${baseUrl}${name}`;
  }

  return `${baseUrl}/${name}`;
}

/* istanbul ignore file */
function duplicateLogger(root, existPath, currentPath, api) {
  console.log('[%s] Duplicate routes detected. %s and %s both resolve to %s.', chalk__default['default'].yellow('warn'), chalk__default['default'].cyan(path.relative(root, existPath)), chalk__default['default'].cyan(path.relative(root, currentPath)), chalk__default['default'].cyan(api));
}

class HTTPRouter extends FileRouter {
  constructor(...args) {
    super(...args);
    this.routes = new Map();
  }

  getBaseUrl(file) {
    const url = this.getHTTPPath(file, '', true);

    if (url === '/*') {
      return '/';
    }

    if (url.endsWith('/*')) {
      return url.slice(0, -2);
    }

    return url;
  }

  getHTTPPath(file, method, isExportDefault) {
    const {
      basePath,
      baseDir,
      underscore
    } = this.getRoute(file);
    const lambdaDirectory = this.getApiDirectory(baseDir);
    const {
      isCatchAllRoutes,
      filename
    } = this.parseFilename(path.basename(file, path.extname(file)));
    const fileRoute = filename === 'index' ? '' : filename; // TODO remove underscore support in future

    const func = isExportDefault ? '' : `${underscore ? '_' : ''}${method}`;
    const api = path.toUnix(path.join(basePath,
    /**
     * /apis/lambda/index.ts -> ''
     * /apis/lambda/note/index.ts -> 'note'
     */
    path.relative(lambdaDirectory, path.dirname(file)),
    /**
     * index -> ''
     * demo -> '/demo'
     */
    fileRoute, // underscore: getNoteList -> _getNoteList
    func, isCatchAllRoutes ? '/*' : ''));
    /**
     * duplicate routes
     */

    const existPath = this.routes.get(api);

    if (existPath && existPath !== path.toUnix(file)) {
      duplicateLogger(this.root, existPath, file, api);
    }

    this.routes.set(api, path.toUnix(file));
    return api;
  }

  parseFilename(filename) {
    var _re$exec;

    const re = /\[\.{3}(.+)]/;
    const isCatchAllRoutes = re.test(filename);
    return {
      isCatchAllRoutes,
      filename: isCatchAllRoutes ? (_re$exec = re.exec(filename)) === null || _re$exec === void 0 ? void 0 : _re$exec[1] : filename
    };
  }

}

class HTTPGateway {
  static is(route) {
    return !!(route !== null && route !== void 0 && route.basePath);
  }

  constructor(options) {
    this.options = void 0;
    this.container = void 0;
    this.router = void 0;

    this.handleError = async next => {
      try {
        await next();
      } catch (error) {
        const ctx = useContext();

        if (this.options.projectConfig.superjson) {
          ctx.status = 500;
          ctx.body = superjson__default['default'].serialize(error);
        } else {
          throw error;
        }
      }
    };

    this.options = options;
    this.router = new HTTPRouter({
      root: this.options.root,
      projectConfig: this.options.projectConfig,
      useSourceFile: this.options.router.useSourceFile
    });
  }

  createApi(options) {
    const {
      file,
      functionName,
      isExportDefault,
      fn
    } = options;
    const httpPath = this.router.getHTTPPath(file, functionName, isExportDefault);
    fn._param.url = httpPath;
    this.createHTTPApi(httpPath, options);
  }

  createHTTPApi(httpPath, options) {
    const {
      functionId,
      fn
    } = options; // setup middleware

    const middleware = [...this.getMiddleware(), ...fn.middleware];
    const FunctionContainer = createFunctionContainer({
      isHTTP: true,
      fn,
      functionId,

      parseArgs(ctx) {
        var _ctx$request, _ctx$request$body;

        const args = ((_ctx$request = ctx.request) === null || _ctx$request === void 0 ? void 0 : (_ctx$request$body = _ctx$request.body) === null || _ctx$request$body === void 0 ? void 0 : _ctx$request$body.args) || [];
        return args;
      },

      classDecorators: [decorator.Controller('/')],
      handlerDecorators: [decorator.All(httpPath, {
        middleware
      })]
    });
    this.container.bind(functionId, FunctionContainer);
  }

  getMiddleware() {
    const mws = [this.handleError];
    const enableSuperjson = this.options.projectConfig.superjson;

    if (enableSuperjson) {
      mws.push(this.deserializeSuperjson);
    }

    return mws.map(useHooksMiddleware);
  }

  async deserializeSuperjson(next) {
    await next();
    const ctx = useContext();

    if (ctx.type.includes('application/json')) {
      ctx.body = superjson__default['default'].serialize(ctx.body);
    }
  }

  afterCreate() {
    if (isDevelopment()) {
      return;
    }

    if (!this.isViteProject) {
      return;
    }

    const {
      routes
    } = this.router;

    if (routes.has('/') || routes.has('/*')) {
      return;
    }

    const baseDir = this.container.get('baseDir');
    const mw = staticCache__default['default']({
      dir: path.join(baseDir, '..', this.options.projectConfig.build.viteOutDir),
      dynamic: true,
      alias: {
        '/': 'index.html',

        /**
         * Add alias for windows, '/' -> '\\'
         * https://github.com/koajs/static-cache/blob/master/index.js#L45
         */
        '\\': 'index.html'
      },
      buffer: true,
      gzip: true
    });

    const fn = async () => {};

    fn.middleware = [mw];
    this.createHTTPApi('/*', {
      functionId: 'hooks-host',
      fn
    });
  }

  get isViteProject() {
    return ['vite.config.ts', 'vite.config.js'].some(config => fs.existsSync(path.join(this.options.root, config)));
  }

}
HTTPGateway.router = HTTPRouter;
HTTPGateway.createApiClient = createHTTPApiClient;

async function createEventApiClient(file, code, router) {
  await esModuleLexer.init;
  const [, exports] = esModuleLexer.parse(code);

  if (exports.length === 0) {
    return null;
  }

  const {
    request: {
      client
    }
  } = router.projectConfig;
  const funcs = exports.map(name => {
    const isExportDefault = name === 'default';
    return {
      functionId: router.getFunctionId(file, name, isExportDefault),
      isExportDefault: isExportDefault,
      name: isExportDefault ? '$default' : name
    };
  });
  return formatCode(buildClient(client, funcs));
}

function buildClient(client, funcs) {
  const buildFunction = func => {
    return `
      export ${func.isExportDefault ? 'default' : ''} function ${func.name} (...args) {
        return request({
          functionId: '${func.functionId}',
          data: { args }
        })
      }
    `;
  };

  const functions = funcs.map(buildFunction).join('\n');
  return `
    import { request } from '${client}';
    ${functions}
  `;
}

class EventGateway {
  constructor() {
    this.container = void 0;
  }

  static is(route) {
    return !!(route !== null && route !== void 0 && route.event);
  }

  createApi(options) {
    const {
      functionId,
      fn,
      route
    } = options;
    const FunctionContainer = createFunctionContainer({
      fn,
      functionId,

      parseArgs(_, event) {
        let args;

        switch (route.event) {
          case 'wechat-miniapp':
            args = event === null || event === void 0 ? void 0 : event.args;
        }

        return args || [];
      },

      handlerDecorators: [decorator.ServerlessTrigger(decorator.ServerlessTriggerType.EVENT, {
        functionName: functionId
      })]
    });
    this.container.bind(functionId, FunctionContainer);
  }

}
EventGateway.router = FileRouter;
EventGateway.createApiClient = createEventApiClient;

//   constructor(actual: any, name: string, reason = 'is invalid') {
//     const message = `[ERR_INVALID_ARG_VALUE]: The argument '${name}' ${reason} ${getTypeMessage(
//       actual
//     )}`
//     super(message)
//   }
// }

class ERR_INVALID_ARG_TYPE extends Error {
  constructor(name, expected, actual) {
    const message = `[ERR_INVALID_ARG_TYPE]: The '${name}' argument must be of type ${expected}${getTypeMessage(actual)}`;
    super(message);
  }

} // fork from https://github.com/nodejs/node/blob/master/lib/internal/errors.js

function getTypeMessage(actual) {
  let msg = '';

  if (actual == null) {
    msg += `. Received ${actual}`;
  } else if (typeof actual === 'function' && actual.name) {
    msg += `. Received function ${actual.name}`;
  } else if (typeof actual === 'object') {
    if (actual.constructor && actual.constructor.name) {
      msg += `. Received an instance of ${actual.constructor.name}`;
    } else {
      const inspected = util.inspect(actual, {
        depth: -1
      });
      msg += `. Received ${inspected}`;
    }
  } else {
    let inspected = util.inspect(actual, {
      colors: false
    });
    if (inspected.length > 25) inspected = `${inspected.slice(0, 25)}...`;
    msg += `. Received type ${typeof actual} (${inspected})`;
  }

  return msg;
}

function validateArray(value, name) {
  if (!Array.isArray(value)) {
    throw new ERR_INVALID_ARG_TYPE(name, 'Array', value);
  }
}
function validateString(value, name) {
  if (typeof value !== 'string') throw new ERR_INVALID_ARG_TYPE(name, 'string', value);
} // export function validateOneOf(value: any, name: string, oneOf: any[]) {
//   if (!oneOf.includes(value)) {
//     const allowed = oneOf
//       .map((v) => (typeof v === 'string' ? `'${v}'` : String(v)))
//       .join(', ')
//     const reason = 'must be one of: ' + allowed
//     throw new ERR_INVALID_ARG_VALUE(name, value, reason)
//   }
// }

function validateFunction(value, name) {
  if (typeof value !== 'function') {
    throw new ERR_INVALID_ARG_TYPE(name, 'Function', value);
  }
}

function getProjectRoot(cwd) {
  return pkgDir.sync(cwd) || process.cwd();
}
const ignorePattern = req => {
  // Ignore Vite dev server
  const vite = ['@vite', '@react-refresh'];

  if (vite.some(api => req.url.includes(api))) {
    return true;
  }

  const url = require('url');

  const {
    pathname,
    query
  } = url.parse(req.url);
  const reg = /\.(js|ts|tsx|css|less|sass|scss|map|json|png|jpg|jpeg|gif|svg|eot|woff2|ttf)$/;
  return reg.test(pathname) || reg.test(query);
};
function getConfig(cwd) {
  const root = getProjectRoot(cwd);
  const configs = {
    ts: path__default['default'].join(root, 'midway.config.ts'),
    js: path__default['default'].join(root, 'midway.config.js')
  };

  if (!fs.existsSync(configs.ts) && !fs.existsSync(configs.js)) {
    throw new Error(`[ERR_INVALID_CONFIG] midway.config.ts is not found, root: ${root}`);
  }

  const userConfig = tryRequire(configs.ts) || tryRequire(configs.js);
  const builtinGateways = new Set();

  for (const route of userConfig.routes) {
    if (route.basePath) builtinGateways.add(HTTPGateway);
    if (route.event) builtinGateways.add(EventGateway);
  }

  userConfig.gateway = userConfig.gateway || [];
  userConfig.gateway.push(...builtinGateways);
  return ___default['default'].defaultsDeep({}, userConfig, {
    superjson: false,
    source: './src/apis',
    dev: {
      ignorePattern
    },
    build: {
      viteOutDir: './build',
      outDir: './dist'
    },
    request: {
      client: '@midwayjs/hooks-core/request'
    }
  });
}
function defineConfig(config) {
  validateArray(config.routes, 'defineConfig.routes');
  return config;
}

const tryRequire = id => {
  try {
    const jiti = createJITI__default['default']();
    const contents = jiti(id);
    if ('default' in contents) return contents.default;
    return contents;
  } catch {
    return undefined;
  }
};

class HooksComponent {
  constructor(options) {
    this.options = void 0;
    this.adapters = void 0;

    this.useAsyncLocalStorage = async (ctx, next) => {
      await als.run({
        ctx
      }, async () => {
        await next();
      });
    };

    this.options = options;
    this.adapters = options.projectConfig.gateway.map(Adapter => new Adapter(this.options));
  }

  init() {
    const {
      router,
      projectConfig: {
        routes
      }
    } = this.options;
    let count = 0;
    const totalCount = routes.reduce((totalCount, route) => {
      // Windows
      const dir = path__default['default'].join(router.source, route.baseDir);
      const files = globby.sync([dir]).filter(file => router.isApiFile(file));
      return totalCount + files.length;
    }, 0);
    const configuration = core.createConfiguration({
      namespace: '@midwayjs/hooks',
      directoryResolveFilter: routes.map(route => {
        return {
          // Windows
          pattern: path__default$1['default'].join(router.source, route.baseDir),
          ignoreRequire: true,
          filter: (_, file, container) => {
            if (!router.isApiFile(file)) {
              return;
            } // Initialize container


            this.adapters.forEach(adapter => adapter.container = container); // Create api function

            const adapter = this.getAdapterByRoute(route);

            const mod = require(file);

            this.createApi(mod, adapter, file, route);
            container.bindClass(mod, '', file); // Call afterCreate hooks

            count++;

            if (count === totalCount) {
              this.adapters.forEach(adapter => {
                var _adapter$afterCreate;

                return (_adapter$afterCreate = adapter.afterCreate) === null || _adapter$afterCreate === void 0 ? void 0 : _adapter$afterCreate.call(adapter);
              });
            }
          }
        };
      })
    });
    configuration.onReady((_, app) => {
      // Setup global middleware
      for (const mw of this.getGlobalMiddleware()) {
        app.use(mw);
      }
    }).onStop(___default['default'].noop);
    return {
      Configuration: configuration
    };
  }

  getAdapterByRoute(route) {
    const Adapter = this.options.router.getGatewayByRoute(route);
    return this.adapters.find(inst => inst instanceof Adapter);
  }

  createApi(mod, adapter, file, route) {
    var _mod$config;

    const modMiddleware = (mod === null || mod === void 0 ? void 0 : (_mod$config = mod.config) === null || _mod$config === void 0 ? void 0 : _mod$config.middleware) || [];

    const funcs = ___default['default'].pickBy(mod, ___default['default'].isFunction);

    for (const [name, fn] of Object.entries(funcs)) {
      fn.middleware = (fn.middleware || (fn.middleware = [])).concat(modMiddleware).map(useHooksMiddleware);
      const isExportDefault = name === 'default';
      const functionName = isExportDefault ? '$default' : name;
      const functionId = this.options.router.getFunctionId(file, functionName, isExportDefault);
      fn._param = {
        functionId
      };
      adapter.createApi({
        fn,
        functionName,
        isExportDefault,
        functionId,
        file,
        route
      });
    }
  }
  /**
   * @description Web only
   */


  getGlobalMiddleware() {
    var _this$options$runtime, _this$options$runtime2;

    const mws = [this.useAsyncLocalStorage];
    (_this$options$runtime = this.options.runtimeConfig.middleware) === null || _this$options$runtime === void 0 ? void 0 : (_this$options$runtime2 = _this$options$runtime.forEach) === null || _this$options$runtime2 === void 0 ? void 0 : _this$options$runtime2.call(_this$options$runtime, mw => mws.push(useHooksMiddleware(mw)));
    return mws;
  }

}

/**
 * @description Create hooks component
 */

const hooks = (runtimeConfig = {}) => {
  if (runtimeConfig.middleware !== undefined) {
    validateArray(runtimeConfig.middleware, 'runtimeConfig.middleware');
  }

  const root = getProjectRoot();
  const projectConfig = getConfig();
  const router = new FileRouter({
    root,
    projectConfig,
    useSourceFile: isDevelopment()
  });
  const hooks = new HooksComponent({
    root,
    router,
    runtimeConfig,
    projectConfig
  });
  return hooks.init();
};

function useLogger() {
  const ctx = useContext();
  return ctx.logger;
}
function usePlugin(key) {
  validateString(key, 'key');
  const ctx = useContext();
  return ctx.app[key] || ctx[key];
}
async function useInject(identifier) {
  if (typeof identifier !== 'function' && typeof identifier !== 'string') {
    throw new ERR_INVALID_ARG_TYPE('identifier', 'class, string', identifier);
  }

  const ctx = useContext();
  const requestContext = ctx['requestContext'];
  return requestContext.getAsync(identifier);
}
function useConfig(key) {
  if (key !== undefined) {
    validateString(key, 'key');
  }

  const ctx = useContext();
  const requestContext = ctx['requestContext'];
  return requestContext.getConfigService().getConfiguration(key);
}

Object.defineProperty(exports, 'superjson', {
  enumerable: true,
  get: function () {
    return superjson__default['default'];
  }
});
Object.defineProperty(exports, 'consola', {
  enumerable: true,
  get: function () {
    return consola__default['default'];
  }
});
exports.defaults = request_dist_midwayjsHooksCoreRequest.defaults;
exports.request = request_dist_midwayjsHooksCoreRequest.request;
exports.ERR_INVALID_ARG_TYPE = ERR_INVALID_ARG_TYPE;
exports.EventGateway = EventGateway;
exports.FileRouter = FileRouter;
exports.HTTPGateway = HTTPGateway;
exports.HTTPRouter = HTTPRouter;
exports.als = als;
exports.createEventApiClient = createEventApiClient;
exports.createFunctionContainer = createFunctionContainer;
exports.createHTTPApiClient = createHTTPApiClient;
exports.defineConfig = defineConfig;
exports.formatCode = formatCode;
exports.getConfig = getConfig;
exports.getProjectRoot = getProjectRoot;
exports.hooks = hooks;
exports.ignorePattern = ignorePattern;
exports.isDevelopment = isDevelopment;
exports.useConfig = useConfig;
exports.useContext = useContext;
exports.useHooksMiddleware = useHooksMiddleware;
exports.useInject = useInject;
exports.useLogger = useLogger;
exports.usePlugin = usePlugin;
exports.validateArray = validateArray;
exports.validateFunction = validateFunction;
exports.validateString = validateString;
